function [R,pmax] = normal_regret(N,h)
% Code to accompany "Trial size for near-optimal choice between
% surveillance and aggressive treatment: reconsidering MSLT-II"
% by Charles F. Manski and Aleksey Tetenov
% October 2018
% We thank Michael Gmeiner for helpful comments regarding the code.
%
% [R,pmax] = normal_regret(N,h)
%
% This function computes the normal approximation
% to maximum regret of the Empirical Success treatment rule
% for trials of two treatments.
% Treatment A has a binary outcome y(A) in {0,1}.
% Treatment B has a binary main outcome y(B) in {0,1}
% and a binary side effect s(B) in {0,1}.
% The maximization problem is described in Appendix A.2
%
% Inputs:
% N: sample size per arm (positive scalar)
% h: harm to the patient if s(B) = 1 (non-negative scalar)
% 0 <= h
%
% Outputs:
% R: maximum regret (normal approximation)
% pmax: 1x4 vector of parameters of the distribution of treatment response 
% that maximize the normal regret approximation of regret:
%   pmax(1) = P(y(A) = 1), main effect of treatment A
%   pmax(2) = P(y(B) = 1), main effect of treatment B
% conditional probabilities of side effects of treatment B:
%   pmax(3) = P(s(B) = 1|y(B) = 0) = 1
%   pmax(4) = P(s(B) = 1|y(B) = 1) = 0
%
% Tested using MATLAB 2018b
% Requires Optimization and Statistics toolboxes

% Check inputs
if ~(isscalar(N) && N >= 1)
	error('First argument N must be a positive scalar N>=1')
end
if ~(isscalar(h) && h >= 0)
	error('Second argument h must be a scalar h>=0')
end

R = -Inf;

% Options for fmincon
fminoptions = optimoptions('fmincon','Algorithm','sqp','Display','none','StepTolerance',1e-12,'OptimalityTolerance',1e-12,'MaxFunctionEvaluations',Inf,'MaxIterations',100000);

% Parameters: a, b (b*_10 in Appendix A.2)
% The treatment effect is
% tau = (1+h)*b - a - h   (A.3 in the Appendix)
% The variance is
% V = (1-b)*b*(1+h)^2 + a(1-a)   (A.2 in the Appendix)

% Optimization needs to be done over three segments
% given by equations (A.4), (A.5) and (A.6) in the Appendix.
% Optimization over (A.4) and (A.5) is further split into two intervals
% because the functional form of regret approximation (A.1) changes 
% at the point where treatment effect tau=0.

% Segment 1 (A.4)
% parametrized by b:
% a = 0, (1 + h/2)/(1+h) <= b <= 1
% Given that a = 0, (A.2) and (A.3) equal
% tau = (1+h)*b - h
% V   = (1-b)*b*(1+h)^2
% If h = 0, this segment is degenerate
% If h <= 2, then tau>=0 on the whole segment
% If h > 2, then tau<=0 for (1 + h/2)/(1+h) <= b <= h/(1+h)
%            and tau>=0 for h/(1+h) <= b <= 1
% Since the regret approximation (A.1) changes at tau=0, 
% optimization is split into two parts when h > 2.

if h>0
    % Optimization over the segment with tau>=0
    % a=0, max( (1 + h/2)/(1+h), h/(1+h) ) <= b <= 1
    % This segment is not empty for any h>0

    % Define an anonymous function to minimize (-regret)
    rfunc11 = @(b) -((1+h)*b-h)*normcdf(-((1+h)*b-h)*sqrt(N/(b*(1-b)*(1+h)^2)));
    lb11 = max( (1+h/2)/(1+h), h/(1+h) );
    ub11 = 1;
    % Results show that maximum regret is obtained when a is in a 1/sqrt(N)
    % neighborhood of the point where tau = 0, which is h/(1+h) if h>2
    % The starting point is chosen in such a neighborhood. 
    % For large values of N normcdf() and  the regret function
    % evaluate exactly as zero on a large part of the interval.
    % This often prevents the optimization algorithm from leaving
    % the starting point if the starting point has very low regret
    % that is evaluated as zero.
    start11 = lb11 + 0.5*(ub11-lb11)/sqrt(N);
    [b11 v11] = fmincon(rfunc11,start11,[],[],[],[],lb11,ub11,[],fminoptions);
    if -v11>R
        % If maximum regret found on this segment is higher than previous
        % maximum, save its value and maximizing parameters
        R = -v11;
        pmax = [0 b11 1 0];
    end
    
    if h>2
        % Optimization over the segment with tau<=0
        % a=0, (1 + h/2)/(1+h) <= b <= h/(1+h)
        % This segment is empty for h<=2
        
        % Define an anonymous function to minimize (-regret)
        rfunc12 = @(b) ((1+h)*b-h)*normcdf(((1+h)*b-h)*sqrt(N/(b*(1-b)*(1+h)^2)));
        lb12 = (1+h/2)/(1+h);
        ub12 = h/(1+h);
        
        % Results show that maximum regret is obtained when a is in a 1/sqrt(N)
        % neighborhood of the point where tau = 0, which is h/(1+h)
        % The starting point is chosen in such a neighborhood. 
        % For large values of N normcdf() and  the regret function
        % evaluate exactly as zero on a large part of the interval.
        % This often prevents the optimization algorithm from leaving
        % the starting point if the starting point has very low regret
        % that is evaluated as zero.
        start12 = ub12 - 0.5*(ub12-lb12)/sqrt(N);
        [b12 v12] = fmincon(rfunc12,start12,[],[],[],[],lb12,ub12,[],fminoptions);
        if -v12>R
            % If maximum regret found on this segment is higher than previous
            % maximum, save its value and maximizing parameters
            R = -v12;
            pmax = [0 b12 1 0];
        end
    end
end



% Segment 2 (A.5)
% parametrized by a
% 0 <= a <= 1, b = (1+ h/2 - a)/(1+h)
% Substituting the expression for b as a funciton of a, (A.2, A.3) equal
% tau = 1-h/2-2*a
% V   = (1+h/2-a)*(h/2+a) + a*(1-a)
% If h >= 2, then tau<=0 on the whole segment
% If h < 2, then tau>=0 for 0 <= a <= 1/2-h/4
%            and tau<=0 for 1/2-h/4 <= a <= 1
% Since the regret approximation (A.1) changes at tau=0, 
% optimization is split into two parts when h < 2.

if h<2
    % Optimization over the segment with tau>=0
    % 0 <= a <= 1/2-h/4, b = (1+ h/2 - a)/(1+h)
    % This segment is empty for h>=2

    % Define an anonymous function to minimize (-regret)
    rfunc21 = @(a) -(1-h/2-2*a)*normcdf(-(1-h/2-2*a)*sqrt(N/((1+h/2-a)*(h/2+a) + a*(1-a))));
    lb21 = 0;
    ub21 = 1/2 - h/4;
    % Starting point in a 1/sqrt(N) neighborhood of the upper bound
    % For h<2, tau=0 at the upper bound.
    start21 = ub21 - 0.5*(ub21-lb21)/sqrt(N);
    [a21 v21] = fmincon(rfunc21,start21,[],[],[],[],lb21,ub21,[],fminoptions);
    if -v21>R
        % If maximum regret found on this segment is higher than previous
        % maximum, save its value and maximizing parameters
        R = -v21;
        pmax = [a21 (1+ h/2 - a21)/(1+h) 1 0];
    end
end

% Optimization over the segment with tau<=0
% max(0,1/2-h/4) <= a <= 1, b = (1+ h/2 - a)/(1+h)
% This segment is not empty for any h>=0

% Define an anonymous function to minimize (-regret)
rfunc22 = @(a) (1-h/2-2*a)*normcdf((1-h/2-2*a)*sqrt(N/((1+h/2-a)*(h/2+a) + a*(1-a))));
lb22 = max(0,1/2 - h/4);
ub22 = 1;
% Starting point in a 1/sqrt(N) neighborhood of the lower bound
% For h<2, tau=0 at the lower bound.
start22 = lb22 + 0.5*(ub22-lb22)/sqrt(N);
[a22 v22] = fmincon(rfunc22,start22,[],[],[],[],lb22,ub22,[],fminoptions);
if -v22>R
    % If maximum regret found on this segment is higher than previous
    % maximum, save its value and maximizing parameters
    R = -v22;
    pmax = [a22 (1+ h/2 - a22)/(1+h) 1 0];
end

% Segment 3 (A.6):
% parametrized by b
% a = 1, 0 <= b <= (h/2)/(1+h)
% Given that a = 1, (A.2) and (A.3) equal
% tau = (b-1)*(1+h)
% V   = b*(1-b)*(1+h)^2
% If h = 0, this segment is degenerate
% If h > 0, then tau > 0 on the whole segment

if h>0
    % Define an anonymous function to minimize (-regret)
    rfunc3 = @(b) ((b-1)*(1+h))*normcdf(((b-1)*(1+h))*sqrt(N/(b*(1-b)*(1+h)^2)));
    lb3 = 0;
    ub3 = (h/2)/(1+h);
    % Starting point in a 1/sqrt(N) neighborhood of the lower bound
    start3 = lb3 + 0.5*(ub3-lb3)/sqrt(N);
    [b3 v3] = fmincon(rfunc3,start3,[],[],[],[],lb3,ub3,[],fminoptions);
    if -v3>R
        % If maximum regret found on this segment is higher than previous
        % maximum, save its value and maximizing parameters
        R = -v3;
        pmax = [1 b3 1 0];
    end
end

end