function [R,pmax] = regret(N,h)
% Code to accompany "Trial size for near-optimal choice between
% surveillance and aggressive treatment: reconsidering MSLT-II"
% by Charles F. Manski and Aleksey Tetenov
% October 2018
% We thank Michael Gmeiner for helpful comments regarding the code.
%
% [R,pmax] = regret(N,h)
%
% This function computes maximum regret of the Empirical Success 
% treatment rule for trials of two treatments using simulated annealing.
% Treatment A has a binary outcome y(A) in {0,1}.
% Treatment B has a binary main outcome y(B) in {0,1}
% and a binary side effect s(B) in {0,1}.
%
% Inputs:
% N: sample size per arm (positive integer)
% h: harm to the patient if s(B) = 1 (non-negative scalar)
% 0 <= h 
%
% Outputs:
% R: maximum regret
% pmax: 1x4 vector of parameters of the distribution of treatment response 
% that maximize the normal regret approximation of regret:
%   pmax(1) = P(y(A) = 1), main effect of treatment A
%   pmax(2) = P(y(B) = 1), main effect of treatment B
% conditional probabilities of side effects of treatment B:
%   pmax(3) = P(s(B) = 1|y(B) = 0) 
%   pmax(4) = P(s(B) = 1|y(B) = 1) 
%
% Tested using MATLAB 2018b
% Requires Global Optimization, Optimization, and Statistics toolboxes
%
% Runtime (using a desktop with a 4-core i7-6700K CPU)
% with the number of simulated annealing iterations set to
% MaxIter=10000 (this number of iterations is sufficient to replicate
% most entries in Table 1. Replicating the results for N=50,h=0.1 and
% N=70,h=0.1 requires 200000 iterations).
% <  9 sec for N<= 250, h = 0
% ~  1 min for N =  10, h > 0
% ~  6 min for N = 100, h > 0
% ~ 20 min for N = 250, h > 0

% Check inputs
if ~(isscalar(N) && (N==floor(N)) && N > 0)
	error('First argument N must be a positive integer')
end
if ~(isscalar(h) && h >= 0)
	error('Second argument h must be a scalar 0<=h')
end

% Overview:
% Given a 4-dimensional parameter vector p,
% the outcomes of a two-armed trial with N subjects per arm are
% from arm A:
% sum(y(A)) ~ binomial(N,p(1))
% from arm B:
% sum(y(B)) ~ binomial(N,p(2))
% sum(s(B)) = binomial(N-sum(y(B)),p(3)) + binomial(sum(y(B)),p(4))
%
% The program searches over all possible values of p in [0,1]^4
% for the value pmax that maximizes regret of the empirical success rule.
% The ES rule chooses: 
% treatment B if sum(y(B)) - h*sum(s(B)) > sum(y(A)),
% treatment A if sum(y(B)) - h*sum(s(B)) < sum(y(A)),
% A with probability 1/2 and B with probability 1/2 (tie-breaking)
% if sum(y(B)) - h*sum(s(B)) = sum(y(A)).
% The ES treatment rule and its regret are discontinuous in harm value h
% because these inequalities are discontinuous in h for a given outcome.
% For a given value of h, though, regret is continuous in parameters p.
%
% Internal function regret_full(p) explicitly calculates the probabilities
% of all possible experimental outcomes and the resulting regret
% of the ES rule. The simulated annealing algorithm is then used to
% maximize this function over the parameter space.
%
% For values of 0<=h<=0.5 considered in the paper, the worst-case values
% of p(3) and p(4) are often 0 or 1.
% To make search on the boundaries of the parameter space more reliable,
% simulated annealing search is also executed separately with 
% p(3) and/or p(4) restricted to be 0 or 1.
% Regret for these boundary searches is computed using internal functions
% regret_0a(p) for p(3) = 0 (these functions take a 3-dimensional input)
% regret_1a(p) for p(3) = 1
% regret_a0(p) for p(4) = 0
% regret_a1(p) for p(4) = 1
% regret_00(p) for p(3) = 0 and p(4) = 0 (2-dimensional input)
% regret_01(p) for p(3) = 0 and p(4) = 1
% regret_10(p) for p(3) = 1 and p(4) = 0
% regret_11(p) for p(3) = 1 and p(4) = 1
% These functions are implemented separately (instead of using regret_full
% with parameter restrictions) because they are optimized to compute
% the probability distribution of outcomes faster with these restrictions.



% Optimization options:
% The function uses simulated annealing, then the solution from
% the simulated annealing step is polished off using fmincon

% Max number of iterations for simulated annealing controls the length
% of search. 10000 is sufficient for most results reported in Table 1,
% but a larger number of iterations is needed in some cases (particularly
% for 0<h<=.1).
MaxIter = 10000;
fminoptions = optimoptions('fmincon','Algorithm','sqp','Display','off','OptimalityTolerance',1e-12,'StepTolerance',1e-12,'MaxFunctionEvaluations',Inf,'MaxIterations',10000);
saoptions = optimoptions(@simulannealbnd,'Display','off','FunctionTolerance',1e-12,'HybridFcn',{@fmincon,fminoptions},'MaxFunctionEvaluations',Inf,'MaxIterations',MaxIter,'MaxStallIterations',MaxIter);



nn = [0:N];
% D00 is the treatment rule (probability of assigning treatment B)
% for all combinations of sum(y(A)) and sum(y(B)) and p(3)=p(4)=0
% It does not change with parameters p, hence is pre-computed outside
% of the regret_00 function for efficiency.
D00 = 0.5*bsxfun(@le, nn', nn) + 0.5*bsxfun(@lt, nn', nn);
% regret_00 computes regret with the restriction that p(3)=p(4)=0
function regret = regret_00(p) 
    % uses 2-dimensional input p
    % p(1) = P(y(A)=1)
    % p(2) = P(y(B)=1)
    % p(3) = P(s(B)=1 | y(B) = 0) = 0
    % P(4) = P(s(B)=1 | y(B) = 1) = 0
    % tau = treatment effect:
    tau = p(2)-p(1);
    % E = probability of assigning treatment B given the distribution
    % of trial outcomes:
    E = binopdf(nn,N,p(1))*D00*binopdf(nn,N,p(2))';
    regret  = - max(0,tau) + E*tau;
end

% 1. 
% Search boundary solutions with p(3)=0,p(4)=0
% For h = 0, this is the only search necessary
[p00 v00] = simulannealbnd(@regret_00,[.5 .5],zeros(2,1),ones(2,1),saoptions);
% R stores maximum regret and will be updated if parameters
% yielding higher regret are found further on.
R = -v00;
% pmax stores parameters yielding maximum regret and will also be updated
% if higher regret is found further on.
pmax = [p00 0 0];



% The remaining functions are used only if h>0, but they always have to
% be defined because MATLAB does not allow functions to be defined inside
% an if statement.



% The constant t_eps is used further below to correctly compute 
% inequalities for treatment rules. The treatment assignment depends
% on the sign of
% sum(y(B)) - h*sum(s(B)) - sum(y(A))
% In cases when sum(y(B))-h*sum(s(B))-sum(y(A)) equals zero, it may not
% evaluate exactly as zero in floating-point arithmetic.
% To avoid these errors, instead of evaluating whether
% sum(y(B))-h*sum(s(B))-sum(y(A)) == 0
% the computations of treatment rules below will evaluate whether
% sum(y(B))-h*sum(s(B))-sum(y(A)) > -t_eps and
% sum(y(B))-h*sum(s(B))-sum(y(A)) > t_eps.
% If the first inequality holds, but the second one doesn't,
% then treatment B will be assigned with probability 1/2.
t_eps = 1e-10;

% Functions for optimization over 2-parameter subfamilies
% with both p(3) and p(4) taking extreme values 0 or 1
% In these cases the treatment rule is only a function of
% sum(y(A)) and sum(y(B)) because s(B) is deterministic given y(B)

% D01 computes the treatment rule (probability of assigning treatment B)
% for all combinations of sum(y(A)) and sum(y(B)) and p(3)=0, p(4)=1
D01 = 0.5*bsxfun(@lt, nn', (1-h).*nn+t_eps) + 0.5*bsxfun(@lt, nn', (1-h).*nn-t_eps);
function regret = regret_01(p)
    tau = (1-h)*p(2)-p(1);
    E = binopdf(nn,N,p(1))*D01*binopdf(nn,N,p(2))';
    regret  = - max(0,tau) + E*tau;
end

% D10 computes the treatment rule (probability of assigning treatment B)
% for all combinations of sum(y(A)) and sum(y(B)) and p(3)=1, p(4)=0
D10 = 0.5*bsxfun(@lt, nn', (1+h).*nn-N*h+t_eps) + 0.5*bsxfun(@lt, nn', (1+h).*nn-N*h-t_eps);
function regret = regret_10(p)
    tau = (p(2)-1)*h + p(2) - p(1);
    E = binopdf(nn,N,p(1))*D10*binopdf(nn,N,p(2))';
    regret  = - max(0,tau) + E*tau;
end

% D11 computes the treatment rule (probability of assigning treatment B)
% for all combinations of sum(y(A)) and sum(y(B)) and p(3)=1, p(4)=1
D11 = 0.5*bsxfun(@lt, nn', nn-N*h+t_eps) + 0.5*bsxfun(@lt, nn', nn-N*h-t_eps);
function regret = regret_11(p)
    tau = p(2)-h-p(1);
    E = binopdf(nn,N,p(1))*D11*binopdf(nn,N,p(2))';
    regret  = - max(0,tau) + E*tau;
end



% Regret function for optimization over all 4 parameters (regret_full),
% which is defined further below the comments and preliminary computations,
% is written to minimize the number of times that binomial pdf or cdf
% is evaluated for each value of p.
% Each individual in treatment B arm will have one of four outcomes:
% y(B)=1,s(B)=1  y(B)=1,s(B)=0  y(B)=0,s(B)=1  y(B)=0,s(B)=0
% The following code will create a list of all possible multinomial
% outcomes of the trial for treatment B (MNOs).
% j-th row of matrix MNOs contains
% MNOs(j,1) = # of outcomes with y(B)=1,s(B)=1
% MNOs(j,2) = # of outcomes with y(B)=1,s(B)=0
% etc.
% The array Mhats will contain the sum of observed utilities of
% treatment B in the trial.
% The probability of each outcome j is a product of 3 binomial pdfs:
% P(#y(B)=1) x P(#s(B)=1,y(B)=0|#y(B)=0) x P(# s(B)=1,y(B)=1|#y(B)=1)
% Many of these probabilities are repeated across outcomes, so they
% will be computed only once.
% For example, array PDF_YB will contain P(sum(y(B))=n) for n=0..N
% The index array iYB will point to the relevant element of PDF_YB,
% so that PDF_YB(iYB) will create an array that contains the correct
% P(#y(B)=1) for each possible multinomial outcome.
% Similarly, PDF_SB0 will contain probabilities P(#s(B)=1,y(B)=0|#y(B)=0)
% for all feasible pairs of #s(B)=1,y(B)=0 and #y(B)=0.
% The index array iSB0 will map each multinomial outcome to the correct
% conditional probability.
% In the regret_full function, computing
% PDF_YB(iYB).*PDF_SB0(iSB0).*PDF_SB1(iSB1)
% will give a vector of outcome probabilities for all possible
% multinomial outcomes of treatment B.
% For each multinomial outcome of B we will then compute the
% probability that the sum of outcomes of treatment A will be
% strictly smaller or <= than Mhats(j).
% This is also done by creating index arrays iA1 and iA2
% that point to the binomial(N,p(1)) CDF (CDF_YA).
% For multinomial outcome j, the j-th element of
% CDF_YA(iA1) will be the conditional probability that the sum of
% utilities of treatment A will be strictly smaller than the sum of
% utilities of treatment B.
% CDF_YA(iA2) will be the probability that they are <=,
% hence (CDF_YA(iA1)+CDF_YA(iA2))/2 will be the vector of probabilities
% with which the ES treatment rule will select treatment B instead of A
% conditionally on each possible multinomial outcome for treatment B.

% Create a matrix of multinomial outcome values of treatment B:
% [YB=1,SB=1  YB=1,SB=0  YB=0,SB=1  YB=0,SB=0]
% Start with all triplets of values between 0 and N:
[X1,X2,X3] = ndgrid(nn);
X4 = N-X1-X2-X3;
% Select only feasible multinomial outcomes:
select = (X4>=0)&(X4<=N);
MNOs = [X1(select) X2(select) X3(select) X4(select)];
% clear unnecessary arrays from memory
clear select X1 X2 X3 X4;
% Compute sample mean for each multinomial outcome
Mhats = MNOs*[1-h; 1; -h; 0];

% For each outcome of B construct indices
% for # of outcomes of treatment A 
% for which muhatA < muhatB
iA1 = int32(max(1,floor(Mhats-t_eps)+2));
% for which muhatA <= muhatB
iA2 = int32(max(1,floor(Mhats+t_eps)+2));
% Index for the PDF of sum(y(B))
iYB = 1 + MNOs(:,1) + MNOs(:,2);

% All possible parameters for conditional PDFs of s(B):
[X1,X2] = ndgrid(nn);
select = (X1<=X2);
SBns = X1(select);
SBNs = X2(select);
clear X1 X2;

% Reference indexes to precomputed side effect conditional PDFs
iSB1 = int32(iYB.*(iYB-1)/2 + MNOs(:,1) + 1);
% sum(y(B)=0) for each multinomial outcome:
iYB0 = MNOs(:,3) + MNOs(:,4);
iSB0 = int32(iYB0.*(iYB0+1)/2 + MNOs(:,3) + 1);
iYB = int32(iYB);

% regret_full computes regret for any (unrestricted) value of p
function regret = regret_full(p)
    % CDF of outcomes of treatment A (plus an additional zero entry)
    % CDF_YA needs to be a row vector
    CDF_YA = cumsum([0 binopdf(nn,N,p(1))]);
    % PDF of sum(y(B))
    PDF_YB = binopdf(nn',N,p(2));
    % PDFs of conditional distributions for s(B)
    PDF_SB0 = binopdf(SBns,SBNs,p(3));
    PDF_SB1 = binopdf(SBns,SBNs,p(4));
    % treatment effect
    tau = p(2) -h*(p(2)*p(4)+(1-p(2))*p(3)) - p(1);
    % Compute 2P(muhatB > muhatA) + P(muhatB = muhatA)
    E = (CDF_YA(iA1)+CDF_YA(iA2))*(PDF_YB(iYB).*PDF_SB0(iSB0).*PDF_SB1(iSB1));
    regret  = - max(0,tau) + E*tau*.5;
end



% To compute regret more efficiently in boundary cases when either
% p(3) or p(4) takes an extreme value, we modify regret_full to
% exclude all multinomimal outcomes that have zero probability.

% When p(3)=0, the number of (y(B)=0,s(B)=1) outcomes is always zero
% Select only outcome rows where MNOs(:,3) = 0
select_0a = (MNOs(:,3)==0);
iA1_0a = iA1(select_0a);
iA2_0a = iA2(select_0a);
iYB_0a = iYB(select_0a);
iSB1_0a = iSB1(select_0a);
function regret = regret_0a(p)
    % Expand into a 4-dimensional parameter vector
    p = [p(1) p(2) 0 p(3)];
    CDF_YA = cumsum([0 binopdf(nn,N,p(1))]);
    PDF_YB = binopdf(nn',N,p(2));
    % PDFs of conditional distributions for s(B)
    % PDF_SB0 is not needed (degenerate conditional distribution)
    PDF_SB1 = binopdf(SBns,SBNs,p(4));
    % treatment effect
    tau = p(2) - h*(p(2)*p(4)+(1-p(2))*p(3)) - p(1);
    % Compute 2P(muhatB > muhatA) + P(muhatB = muhatA)
    E = (CDF_YA(iA1_0a)+CDF_YA(iA2_0a))*(PDF_YB(iYB_0a).*PDF_SB1(iSB1_0a));
    regret  = - max(0,tau) + E*tau*.5;
end

% When p(3)=1, the number of (y(B)=0,s(B)=0) outcomes is always zero
% Select only outcome rows where MNOs(:,4) = 0
select_1a = (MNOs(:,4)==0);
iA1_1a = iA1(select_1a);
iA2_1a = iA2(select_1a);
iYB_1a = iYB(select_1a);
iSB1_1a = iSB1(select_1a);
function regret = regret_1a(p)
    p = [p(1) p(2) 1 p(3)];
    CDF_YA = cumsum([0 binopdf(nn,N,p(1))]);
    PDF_YB = binopdf(nn',N,p(2));
    % PDF_SB0 is not needed (degenerate conditional distribution)
    PDF_SB1 = binopdf(SBns,SBNs,p(4));
    tau = p(2) - h*(p(2)*p(4)+(1-p(2))*p(3)) - p(1);
    E = (CDF_YA(iA1_1a)+CDF_YA(iA2_1a))*(PDF_YB(iYB_1a).*PDF_SB1(iSB1_1a));
    regret  = - max(0,tau) + E*tau*.5;
end

% When p(4)=0, the number of (y(B)=1,s(B)=1) outcomes is always zero
% Select only outcome rows where MNOs(:,1) = 0
select_a0 = (MNOs(:,1)==0);
iA1_a0 = iA1(select_a0);
iA2_a0 = iA2(select_a0);
iYB_a0 = iYB(select_a0);
iSB0_a0 = iSB0(select_a0);
function regret = regret_a0(p)
    p = [p(1) p(2) p(3) 0];
    CDF_YA = cumsum([0 binopdf(nn,N,p(1))]);
    PDF_YB = binopdf(nn',N,p(2));
    PDF_SB0 = binopdf(SBns,SBNs,p(3));
    % PDF_SB1 is not needed (degenerate conditional distribution)
    tau = p(2) - h*(p(2)*p(4)+(1-p(2))*p(3)) - p(1);
    E = (CDF_YA(iA1_a0)+CDF_YA(iA2_a0))*(PDF_YB(iYB_a0).*PDF_SB0(iSB0_a0));
    regret  = - max(0,tau) + E*tau*.5;
end

% When p(4)=1, the number of (y(B)=1,s(B)=0) outcomes is always zero
% Select only outcome rows where MNOs(:,2) = 0
select_a1 = (MNOs(:,2)==0);
iA1_a1 = iA1(select_a1);
iA2_a1 = iA2(select_a1);
iYB_a1 = iYB(select_a1);
iSB0_a1 = iSB0(select_a1);
function regret = regret_a1(p)
    p = [p(1) p(2) p(3) 1];
    CDF_YA = cumsum([0 binopdf(nn,N,p(1))]);
    PDF_YB = binopdf(nn',N,p(2));
    PDF_SB0 = binopdf(SBns,SBNs,p(3));
    % PDF_SB1 is not needed (degenerate conditional distribution)
    tau = p(2) - h*(p(2)*p(4)+(1-p(2))*p(3)) - p(1);
    E = (CDF_YA(iA1_a1)+CDF_YA(iA2_a1))*(PDF_YB(iYB_a1).*PDF_SB0(iSB0_a1));
    regret  = - max(0,tau) + E*tau*.5;
end



if (h>0) % for h>0 we need to consider other cases
    
    % Search over remaining 2-parameter families
    [p01 v01] = simulannealbnd(@regret_01,[.5 .5],zeros(2,1),ones(2,1),saoptions);
    if -v01>R
        R = -v01;
        pmax = [p01 0 1];
    end
    [p10 v10] = simulannealbnd(@regret_10,[.5 .5],zeros(2,1),ones(2,1),saoptions);
    if -v10>R
        R = -v10;
        pmax = [p10 1 0];
    end
    [p11 v11] = simulannealbnd(@regret_11,[.5 .5],zeros(2,1),ones(2,1),saoptions);
    if -v11>R
        R = -v11;
        pmax = [p11 1 1];
    end
    
    
    
    % Search over 3-parameter families
    [p0a v0a] = simulannealbnd(@regret_0a,[pmax(1) pmax(2) pmax(4)],zeros(3,1),ones(3,1),saoptions);
    if -v0a>R
        R = -v0a;
        pmax = [p0a(1) p0a(2) 0 p0a(3)];
    end
    [p1a v1a] = simulannealbnd(@regret_1a,[pmax(1) pmax(2) pmax(4)],zeros(3,1),ones(3,1),saoptions);
    if -v1a>R
        R = -v1a;  
        pmax = [p1a(1) p1a(2) 1 p1a(3)];
    end
    [pa0 va0] = simulannealbnd(@regret_a0,[pmax(1) pmax(2) pmax(3)],zeros(3,1),ones(3,1),saoptions);
    if -va0>R
        R = -va0;  
        pmax = [pa0 0];
    end
    [pa1 va1] = simulannealbnd(@regret_a1,[pmax(1) pmax(2) pmax(3)],zeros(3,1),ones(3,1),saoptions);
    if -va1>R
        R = -va1;  
        pmax = [pa1 1];
    end
    % Full 4-parameter search
    [pfull vfull] = simulannealbnd(@regret_full,pmax,zeros(4,1),ones(4,1),saoptions);
    if -vfull>R
        R = -vfull;
        pmax = pfull;
    end
end
end